from typing import Any, Dict

from pydantic import BaseModel

from tdm.abstract.json_schema import DocumentMetadataFields, FactMetadataFields


class MetadataModel(BaseModel):

    def to_metadata(self) -> Dict[str, Any]:
        return self.dict(exclude_none=True)


class DocumentMetadataModel(DocumentMetadataFields, MetadataModel):
    pass


class FactMetadataModel(FactMetadataFields, MetadataModel):
    pass



from typing_extensions import Literal

from tdm.abstract.datamodel.directive import DirectiveType
from tdm.datamodel.directive import CreatePlatformDirective
from tdm.json_schema.directive.abstract import AbstractDirectiveModel


class CreatePlatformDirectiveModel(AbstractDirectiveModel):
    key: str
    platform_type: str
    name: str
    url: str

    directive_type: Literal[DirectiveType.CREATE_PLATFORM] = DirectiveType.CREATE_PLATFORM

    @classmethod
    def build(cls, directive: CreatePlatformDirective) -> 'CreatePlatformDirectiveModel':
        return cls.construct(
            key=directive.key,
            name=directive.name,
            url=directive.url,
            platform_type=directive.platform_type
        )

    def to_directive(self) -> CreatePlatformDirective:
        return CreatePlatformDirective(
            key=self.key,
            name=self.name,
            url=self.url,
            platform_type=self.platform_type
        )

    def __hash__(self):
        return hash((self.directive_type, self.key))



__all__ = [
    'CreateAccountDirective',
    'CreateConceptDirective',
    'CreatePlatformDirective'
]

from .create_account import CreateAccountDirective
from .create_concept import CreateConceptDirective
from .create_platform import CreatePlatformDirective



__all__ = [
    'ACCOUNT_CONCEPT', 'PLATFORM_CONCEPT', 'ConceptFact',
    'ACCOUNT_NAME', 'ACCOUNT_URL', 'PLATFORM_NAME', 'PLATFORM_TYPE', 'PLATFORM_URL', 'PropertyFact', 'PropertyLinkValue',
    'ACCOUNT_PLATFORM', 'RelationFact', 'RelationLinkValue',
    'ValueFact'
]

from .concept import ACCOUNT_CONCEPT, ConceptFact, PLATFORM_CONCEPT
from .property import ACCOUNT_NAME, ACCOUNT_URL, PLATFORM_NAME, PLATFORM_TYPE, PLATFORM_URL, PropertyFact, PropertyLinkValue
from .relation import ACCOUNT_PLATFORM, RelationFact, RelationLinkValue
from .value import ValueFact



__all__ = [
    'CreateAccountDirectiveModel', 'CreateConceptDirectiveModel', 'CreatePlatformDirectiveModel'
]

from .create_account import CreateAccountDirectiveModel
from .create_concept import CreateConceptDirectiveModel
from .create_platform import CreatePlatformDirectiveModel



from abc import abstractmethod
from typing import Any, Callable, Dict, Optional, Tuple

from pydantic import BaseModel

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, AbstractTreeDocumentContent, FactStatus, FactType
from tdm.abstract.datamodel.fact import AbstractLinkValue, FactMetadata
from tdm.datamodel import TalismanSpan
from tdm.json_schema.metadata import FactMetadataModel


class SpanModel(BaseModel):
    node_id: str
    value: str
    start: int
    end: int

    def to_span(self) -> TalismanSpan:
        return TalismanSpan(start_idx=self.start, end_idx=self.end, id_=self.node_id)

    @classmethod
    def build(cls, span: AbstractTalismanSpan, doc: AbstractTreeDocumentContent) -> 'SpanModel':
        return cls.construct(node_id=span.node_id, value=doc.text_for(span), start=span.start_idx, end=span.end_idx)


class PropertyLinkValueModel(BaseModel):
    property_id: Optional[str]
    from_fact: str
    to_fact: str

    @classmethod
    def build(cls, value: AbstractLinkValue) -> 'PropertyLinkValueModel':
        return cls.construct(property_id=value.property_id, from_fact=value.from_fact.id, to_fact=value.to_fact.id)


# Callable[[id, status, type id, value, mention, metadata], fact]
FactFactory = Callable[[str, FactStatus, str, Any, Tuple[AbstractTalismanSpan, ...], FactMetadata], AbstractFact]


class AbstractFactModel(BaseModel):
    id: str
    fact_type: FactType
    status: FactStatus
    type_id: str
    value: Any
    mention: Optional[Tuple[SpanModel, ...]]
    metadata: Optional[FactMetadataModel]

    def to_fact_factory(self) -> Tuple[FactType, Callable[[Dict[str, AbstractFact]], AbstractFact]]:
        def create_fact(mapping: Dict[str, AbstractFact]) -> AbstractFact:
            mention = tuple(span_model.to_span() for span_model in self.mention) if self.mention is not None else None
            metadata = self.metadata.to_metadata() if self.metadata is not None else None
            return self.fact_factory(self.id, self.status, self.type_id, self.to_value(mapping), mention, metadata)

        return self.fact_type, create_fact

    @abstractmethod
    def to_value(self, mapping: Dict[str, AbstractFact]) -> Any:
        pass

    @property
    @abstractmethod
    def fact_factory(self) -> FactFactory:
        pass

    @classmethod
    @abstractmethod
    def build_value(cls, value: Any) -> Any:
        pass

    def __hash__(self) -> int:
        return hash((self.id, self.fact_type, self.status, self.type_id))



import uuid
from typing import Iterable, Optional, Tuple

from tdm.abstract.datamodel import AbstractDirective
from tdm.abstract.datamodel.directive import DirectiveType


class CreateConceptDirective(AbstractDirective):
    __slots__ = ('_name', '_concept_type', '_notes', '_markers', '_access_level', '_id')

    def __init__(self, name: str, concept_type: str, notes: Optional[str] = None, markers: Optional[Iterable[str]] = None,
                 access_level: Optional[str] = None, id_: Optional[str] = None):
        super().__init__(DirectiveType.CREATE_CONCEPT)
        self._id = id_ or self.generate_id()
        self._name = name
        self._concept_type = concept_type
        self._notes = notes
        self._markers = tuple(markers) if markers else None
        self._access_level = access_level

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def concept_type(self) -> str:
        return self._concept_type

    @property
    def notes(self) -> Optional[str]:
        return self._notes

    @property
    def markers(self) -> Optional[Tuple[str]]:
        return self._markers

    @property
    def access_level(self):
        return self._access_level

    @staticmethod
    def generate_id() -> str:
        return str(uuid.uuid4())

    def __eq__(self, other):
        if not isinstance(other, AbstractDirective):
            return NotImplemented
        return isinstance(other, CreateConceptDirective) and other._id == self._id and other._name == self._name and \
            other._concept_type == self._concept_type and other._notes == self._notes and other._markers == self._markers and \
            other._access_level == self._access_level

    def __hash__(self):
        return hash((super().__hash__(), self.concept_type, self._id))
